@@ -1,9 +1,9 @@
-// Copyright (C) 2003, 2005 Free Software Foundation, Inc.
+// Copyright (C) 2003, 2005, 2009 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
 // terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2, or (at your option)
+// Free Software Foundation; either version 3, or (at your option)
 // any later version.
 
 // This library is distributed in the hope that it will be useful,
@@ -12,12 +12,13 @@
 // GNU General Public License for more details.
 
 // You should have received a copy of the GNU General Public License along
-// with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
-// USA.
+// with this library; see the file COPYING3.  If not see
+// <http://www.gnu.org/licenses/>.
 
 // 27.8.1.4 Overridden virtual functions
 
+// { dg-require-fileio "" }
+
 #include <locale>
 #include <fstream>
 #include <testsuite_hooks.h>
@@ -38,30 +39,17 @@ void test01()
 	     traits_type::state_type>);
 
   basic_filebuf<pod_uchar> fb;
+  fb.pubsetbuf(0, 0);
   fb.pubimbue(loc);
 
   fb.open(name, ios_base::in | ios_base::out | ios_base::trunc);
-  fb.sputc(pod_uchar::from<char>('b'));
-  fb.sputc(pod_uchar::from<char>(0xff));
-  fb.sputc(pod_uchar::from<char>('c'));
-  fb.sputc(pod_uchar::from<char>(0xff));
-
-  fb.pubseekoff(0, ios_base::beg);
-  fb.sbumpc();
-  fb.sbumpc();
   traits_type::pos_type pos = fb.pubseekoff(0, ios_base::cur);
-  fb.sbumpc();
-  fb.sbumpc();
-
-  // Check that seekoff sets the current state for output
-  fb.pubseekpos(pos);
-
   fb.sputc(pod_uchar::from<char>('a'));
+  fb.sputc(pod_uchar::from<char>(0xff));
   fb.sputc(pod_uchar::from<char>(0));
   fb.sputc(pod_uchar::from<char>(0));
-  fb.sputc(pod_uchar::from<char>(0));
-  fb.sputc(pod_uchar::from<char>(0));
-  
+
+  // Check that seekpos restores the state correctly
   fb.pubseekpos(pos);
 
   traits_type::int_type c = fb.sbumpc();
